


#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>

#include "globdef.h"
#include "uidef.h"
#include "fft1def.h"
#include "fft2def.h"
#include "fft3def.h"
#include "screendef.h"
#include "sigdef.h"
#include "seldef.h"
#include "thrdef.h"
#include "rusage.h"
#include "txdef.h"
#include "options.h"
#include "conf.h"

#if BUFBARS == TRUE
#define RX_INDICATOR_MAXNUM 8
extern int rx_indicator_bufpos[RX_INDICATOR_MAXNUM];
#endif

int baseb_output_block;
int baseb_min_block;

void blocking_rxout(void)
{
#if RUSAGE_OLD == TRUE
int local_workload_counter;
#endif
int local_workload_reset;
char s[80];
char cc;
float sleep_time, short_sleep_time;
#if OSNUM == OS_FLAG_LINUX
clear_thread_times(THREAD_BLOCKING_RXOUT);
#endif
#if RUSAGE_OLD == TRUE
local_workload_counter=workload_counter;
#endif
local_workload_reset=workload_reset_flag;
sleep_time=2000000./snd[RXDA].interrupt_rate;
short_sleep_time=0.2*sleep_time;
thread_status_flag[THREAD_BLOCKING_RXOUT]=THRFLAG_ACTIVE;
lir_await_event(EVENT_BLOCKING_RXOUT);
while(!kill_all_flag && 
         thread_command_flag[THREAD_BLOCKING_RXOUT]==THRFLAG_ACTIVE)
  {
// *******************************************************
// *******************************************************
#if RUSAGE_OLD == TRUE
  if(local_workload_counter != workload_counter)
    {
    local_workload_counter=workload_counter;
    update_snd(RXDA);
    if(snd[RXDA].valid_bytes > snd[RXDA].block_bytes)
      {
      make_thread_times(THREAD_BLOCKING_RXOUT);
      update_snd(RXDA);
      }
    }
#endif
  if(local_workload_reset!=workload_reset_flag)
    {
    update_snd(RXDA);
    snd[RXDA].min_valid_frames=snd[RXDA].tot_frames;
    local_workload_reset=workload_reset_flag;
    }
  if( ((daout_pa-daout_px+daout_size)&daout_bufmask) > snd[RXDA].block_bytes)
    {
    if(audio_dump_flag==0 && rx_audio_out != -1)
      {
      lir_rx_dawrite();
      }
    else
      {
      daout_px=(daout_px+snd[RXDA].block_bytes)&daout_bufmask;
      }
    }
  else
    {
    if(audio_dump_flag == 1 || rx_audio_out == -1)
      {
      lir_sleep(short_sleep_time);
      }
    else
      {
      update_snd(RXDA);
      if(snd[RXDA].valid_bytes > 2*snd[RXDA].block_bytes)
        {
        lir_sleep(sleep_time);
        goto loop_end;;   
        }
      if(snd[RXDA].valid_bytes < snd[RXDA].block_bytes/4)
        {
        if(count_rx_underrun_flag)
          {
          no_of_rx_underrun_errors++;
          sprintf(s,"RX%s%d",underrun_error_msg,no_of_rx_underrun_errors);
          wg_error(s,WGERR_RXOUT);
          }
// clear the buffer
        if(rx_daout_bytes == 1)
          {
          cc=0x80;
          }
        else
          {
          cc=0;
          }  
        daout_px=(daout_px-snd[RXDA].block_bytes+daout_size)&daout_bufmask;
        memset(&daout[daout_px],cc,snd[RXDA].block_bytes);
        lir_rx_dawrite();
        lir_sleep(short_sleep_time);
        }
      else
        {
        lir_sleep(short_sleep_time);
        }
      }
    }  
loop_end:;
  }
thread_status_flag[THREAD_BLOCKING_RXOUT]=THRFLAG_RETURNED;
while(!kill_all_flag &&
            thread_command_flag[THREAD_BLOCKING_RXOUT] != THRFLAG_NOT_ACTIVE)
  {
  lir_sleep(1000);
  }
}

void rx_output(void)
{
#if RUSAGE_OLD == TRUE
int local_workload_counter;
#endif
char delay_margin_flag;
char cc, blocking;
int dasync_time_interval, da_start_bytes;
int dasync_errors;
int daout_py;
int i, k;
char s[80];
double dt1;
int local_channels, local_bytes;
int local_workload_reset;
int speed_cnt;
float local_fft3_blocktime;
float t1, t2;
float sleep_time;
float short_sleep_time;
double total_time2;
#if OSNUM == OS_FLAG_LINUX
clear_thread_times(THREAD_RX_OUTPUT);
#endif
#if RUSAGE_OLD == TRUE
local_workload_counter=workload_counter;
#endif
blocking=TRUE;
#if PORTAUDIO_PRESENT == 1
if( (ui.use_alsa&PORTAUDIO_RX_OUT) != 0)
  {
  blocking=FALSE;
  }
#endif
thread_status_flag[THREAD_RX_OUTPUT]=THRFLAG_ACTIVE;
local_channels=rx_daout_channels;
local_bytes=rx_daout_bytes;
local_fft3_blocktime=fft3_blocktime;
daout_pa=0;
daout_px=0;
count_rx_underrun_flag=FALSE;
sys_func(THRFLAG_OPEN_RX_SNDOUT);
if(rx_audio_out == -1)blocking=FALSE;
if(kill_all_flag)goto da_output_error;
if(blocking && audio_dump_flag == 0 && rx_audio_out >= 0)
  {
  lir_rx_dawrite();
  daout_pa=0;
  daout_px=0;
  }
dasync_errors=0;
baseb_output_block=128;
da_start_bytes=-1;
thread_command_flag[THREAD_BLOCKING_RXOUT]=THRFLAG_NOT_ACTIVE;
if(blocking)
  {
  lir_init_event(EVENT_BLOCKING_RXOUT);
  lir_sched_yield();
  linrad_thread_create(THREAD_BLOCKING_RXOUT);
  }
while(!kill_all_flag && 
          thread_command_flag[THREAD_RX_OUTPUT]==THRFLAG_ACTIVE)
  {
  lir_sleep(3000);
  }
goto stop_output;
resume:;
if(blocking)
  {
  lir_init_event(EVENT_BLOCKING_RXOUT);
  lir_sched_yield();
  linrad_thread_create(THREAD_BLOCKING_RXOUT);
  }
if(lir_status == LIR_SNDOUT_ERROR) 
  {
  wg_error("Soundcard disabled",WGERR_SOUNDCARD);
  }
thread_status_flag[THREAD_RX_OUTPUT]=THRFLAG_ACTIVE;
count_rx_underrun_flag=FALSE;
lir_sched_yield();
sleep_time=1000000./snd[RXDA].interrupt_rate;
short_sleep_time=0.3*sleep_time;
dasync_time_interval=7*fft3_size/timf3_sampling_speed;
if(dasync_time_interval < 7*fft1_size/timf1_sampling_speed)
                       dasync_time_interval = 7*fft1_size/timf1_sampling_speed;
if(genparm[SECOND_FFT_ENABLE]!=0)
  {
  if(dasync_time_interval < 7*fft2_size/timf1_sampling_speed)
                      dasync_time_interval = 7*fft2_size/timf1_sampling_speed;
  }
if(dasync_time_interval < 0.5)dasync_time_interval=0.5;
local_workload_reset=workload_reset_flag;
speed_cnt=snd[RXDA].interrupt_rate;
baseb_output_block=snd[RXDA].block_bytes/(2*da_resample_ratio);
baseb_min_block=1+1/da_resample_ratio;
if(baseb_min_block < baseb_output_block/16)
  {
  baseb_min_block=baseb_output_block/16;
  }
k=0;
await_narrowband_processing:;
if(kill_all_flag) goto da_output_error;
if(thread_command_flag[THREAD_RX_OUTPUT] != THRFLAG_ACTIVE)goto stop_output;
k++;
lir_sleep(2000);
make_timing_info();
if(kill_all_flag) goto da_output_error;
if(thread_command_flag[THREAD_RX_OUTPUT] != THRFLAG_ACTIVE)goto stop_output;
t1=total_wttim-da_wait_time;
if(t1*genparm[DA_OUTPUT_SPEED] > baseb_output_block)
  {
  i=t1*baseband_sampling_speed;
  if(i > baseb_wts)i=baseb_wts;
  t1-=i/baseband_sampling_speed;
  baseb_py=(baseb_py+i+baseband_size)&baseband_mask;
  baseb_wts-=i;
  }
if(t1*baseband_sampling_speed > 6*baseb_min_block || k<2)goto await_narrowband_processing;
baseb_fx=(baseb_py-2+baseband_size)&baseband_mask;
if(da_start_bytes == -1)
  {
  dasync_counter=0;
  dasync_sum=0;
  dasync_avg1=-1.5;
  dasync_avg2=-1.5;
  dasync_avg3=-1.5;
  dasync_time=current_time();
  dasync_avgtime=dasync_time;
  daspeed_time=dasync_time+1;
  }
daout_pa=0;
daout_px=0;
daout_py=0;
snd[RXDA].min_valid_frames=snd[RXDA].tot_frames;
min_daout_samps=baseband_size;
delay_margin_flag=TRUE;
if(blocking)
  {
  lir_set_event(EVENT_BLOCKING_RXOUT);
  }
while(!kill_all_flag &&
                   thread_command_flag[THREAD_RX_OUTPUT] == THRFLAG_ACTIVE)
  {
#if RUSAGE_OLD == TRUE
  if(local_workload_counter != workload_counter)
    {
    local_workload_counter=workload_counter;
    make_thread_times(THREAD_RX_OUTPUT);
    }
#endif
  speed_cnt--;
  if(speed_cnt==0)
    {
    if(delay_margin_flag)
      {
      snd[RXDA].min_valid_frames=snd[RXDA].tot_frames;
      min_daout_samps=baseband_size;
      delay_margin_flag=FALSE;
      }
    if(local_workload_reset!=workload_reset_flag)
      {
      snd[RXDA].min_valid_frames=snd[RXDA].tot_frames;
      local_workload_reset=workload_reset_flag;
      }
    speed_cnt=snd[RXDA].interrupt_rate;
    total_time2=recent_time;
    if(total_time2-daspeed_time > 0)
      {
// Compute the D/A speed once every second.
      if(audio_dump_flag==0)
        {
        daspeed_time=total_time2+1;
        update_snd(RXDA);
        if(da_start_bytes == -1)
          {
          da_start_time=total_time2;
          da_block_counter=0;
          da_start_bytes=snd[RXDA].empty_bytes;
          snd[RXDA].min_valid_frames=snd[RXDA].tot_frames;
          count_rx_underrun_flag=TRUE;
          }
        else
          {
          measured_da_speed=(da_block_counter*snd[RXDA].block_bytes+
                                  snd[RXDA].empty_bytes-da_start_bytes)/
                          (snd[RXDA].framesize*(total_time2-da_start_time));
          }
        }
// Keep track of the error in the frequency ratio between input and output.
// Average the total processing delay and adjust resampling rate to
// keep it constant.
#define DASYNC_MAXCOUNT 15 //Number used to form average
      dt1=current_time();
      if(  diskread_flag == 0 &&
           new_baseb_flag == 0 &&
           dt1-dasync_avgtime > dasync_time_interval )
        {
        dasync_avgtime=dt1;
        make_timing_info();
        if(kill_all_flag) goto da_output_error;
        if(timinfo_flag != 0)
          {
          sprintf(s,"sync:%f ",total_wttim-dasync_avg3);
          lir_text(26,screen_last_line-8,s);
          }
        if(dasync_avg3 >= 0 && fabs(total_wttim-dasync_avg3)>0.5)
          {
          dasync_errors++;
          sprintf(s,"DA SYNC ERRORS %d",dasync_errors);
          wg_error(s,WGERR_DASYNC);
          baseb_reset_counter++;
          da_start_bytes=-1;
          dasync_counter=0; 
          }
        dasync_counter++;
        dasync_sum+=total_wttim;
        if(dasync_counter >= DASYNC_MAXCOUNT)
          {
          dasync_avg1=dasync_sum/dasync_counter;
          dasync_counter=0;
          dasync_sum=0;
          t2=dt1-dasync_time;
          dasync_time=dt1;
          if(dasync_avg2 > 0)
            {
// The drift between input and output after t2 seconds is
// the difference between the two averages.
// Convert to frequency ratio, but use half the error only
// to avoid oscillations.
// Correct for the deviation from the first average, but only 20%
// so we drift slowy towards it if we are off.
            t1= (t2+0.5*(dasync_avg2-dasync_avg1)+
                                   0.2*(dasync_avg3-dasync_avg1))/t2;
            new_da_resample_ratio=da_resample_ratio*t1;
            dasync_avg2=dasync_avg1;
            }
          else
            {
            if(dasync_avg3 <0)
              {
              dasync_avg3=dasync_avg1;
              }
            else
              {
              dasync_avg2=dasync_avg1;
              }
            }
          }
        }
      if(thread_command_flag[THREAD_RX_OUTPUT] != 
                                            THRFLAG_ACTIVE)goto stop_output;
      }
    }
// ***********************************************************
  lir_sched_yield();
  if( ((baseb_pa-baseb_py+baseband_size)&baseband_mask) > baseb_output_block )
    {
    if( ((daout_px-2*snd[RXDA].block_bytes-daout_pa+daout_size)&daout_bufmask) > (daout_size>>1) )
      {
      make_audio_signal();
      }
    else
      {
      lir_sleep(sleep_time);
      }
    }
  else
    {
    lir_await_event(EVENT_BASEB);
    }
  if( (ui.network_flag&NET_RXOUT_BASEBRAW) != 0)
    {
#if NET_BASEBRAW_MODE == WFM_FM_FIRST_LOWPASS || \
    NET_BASEBRAW_MODE == WFM_FM_FIRST_LOWPASS_UPSAMP || \
    NET_BASEBRAW_MODE == WFM_STEREO_LOW || \
    NET_BASEBRAW_MODE == WFM_STEREO_HIGH    
    int *ntbuf;
    while(baseb_pn != baseb_pa && 
     ((basebrawnet_px-basebrawnet_pa+basebrawnet_mask)&basebrawnet_mask) >16)
      {
      ntbuf=(void*)&basebraw_netsend_buffer[basebrawnet_pa];
      ntbuf[0]=floor(1000000000.*baseb_fm_demod_low[baseb_pn]);
      baseb_pn=(baseb_pn+1)&baseband_mask;
      basebrawnet_pa=(basebrawnet_pa+4)&basebrawnet_mask;
      }
#endif
#if NET_BASEBRAW_MODE == WFM_FM_FULL_BANDWIDTH
    int *ntbuf;
    while(baseb_pn != baseb_pa && 
     ((basebrawnet_px-basebrawnet_pa+basebrawnet_mask)&basebrawnet_mask) >16)
      {
      ntbuf=(void*)&basebraw_netsend_buffer[basebrawnet_pa];
      ntbuf[0]=floor(300000000.*baseb_fm_demod[baseb_pn]);
      baseb_pn=(baseb_pn+1)&baseband_mask;
      basebrawnet_pa=(basebrawnet_pa+4)&basebrawnet_mask;
      }
#endif
#if NET_BASEBRAW_MODE == WFM_AM_FIRST_LOWPASS
    int *ntbuf;
    while(baseb_pn != baseb_pa && 
     ((basebrawnet_px-basebrawnet_pa+basebrawnet_mask)&basebrawnet_mask) >16)
      {
      ntbuf=(void*)&basebraw_netsend_buffer[basebrawnet_pa];
      ntbuf[0]=floor(0.2*sqrt(baseb_carrier_ampl[baseb_pn]));
      baseb_pn=(baseb_pn+1)&baseband_mask;
      basebrawnet_pa=(basebrawnet_pa+4)&basebrawnet_mask;
      }
#endif
#if NET_BASEBRAW_MODE == WFM_FM_FIRST_LOWPASS_RESAMP
    int *ntbuf;
    while(fm1_px != fm1_pa && 
     ((basebrawnet_px-basebrawnet_pa+basebrawnet_mask)&basebrawnet_mask) >16)
      {
      ntbuf=(void*)&basebraw_netsend_buffer[basebrawnet_pa];
      ntbuf[0]=floor(fm1_all[fm1_px]);
//        baseb_pn=(baseb_pn+1)&baseband_mask;
      fm1_px=(fm1_px+1)&fm1_mask;
      basebrawnet_pa=(basebrawnet_pa+4)&basebrawnet_mask;
      }
#endif
#if NET_BASEBRAW_MODE == BASEBAND_IQ
    int *ntbuf;
    while(baseb_pn != baseb_pa && 
     ((basebrawnet_px-basebrawnet_pa+basebrawnet_mask)&basebrawnet_mask) >16)
      {
      ntbuf=(void*)&basebraw_netsend_buffer[basebrawnet_pa];
      for(i=0; i<basebraw_ad_channels; i++)
        {
        ntbuf[i]=floor(0.05*baseb_raw[basebraw_ad_channels*baseb_pn+i]);
        }
      baseb_pn=(baseb_pn+1)&baseband_mask;
      basebrawnet_pa=(basebrawnet_pa+4*basebraw_ad_channels)&basebrawnet_mask;
      }
#endif
#if NET_BASEBRAW_MODE == WFM_SYNTHETIZED_FROM_FIRST_LOWPASS || \
    NET_BASEBRAW_MODE == WFM_SUBTRACT_FROM_FIRST_LOWPASS
    int *ntbuf;
    while(baseb_pn != baseb_pa && 
     ((basebrawnet_px-basebrawnet_pa+basebrawnet_mask)&basebrawnet_mask) >16)
      {
      ntbuf=(void*)&basebraw_netsend_buffer[basebrawnet_pa];
      for(i=0; i<basebraw_ad_channels; i++)
        {
        ntbuf[i]=floor(0.05*baseb_carrier[basebraw_ad_channels*baseb_pn+i]);
        }
      baseb_pn=(baseb_pn+1)&baseband_mask;
      basebrawnet_pa=(basebrawnet_pa+4*basebraw_ad_channels)&basebrawnet_mask;
      }
#endif
#if NET_BASEBRAW_MODE == WFM_AM_FULL_BANDWIDTH
    int *ntbuf;
    while(baseb_pn != baseb_pa && 
     ((basebrawnet_px-basebrawnet_pa+basebrawnet_mask)&basebrawnet_mask) >16)
      {
      ntbuf=(void*)&basebraw_netsend_buffer[basebrawnet_pa];
      ntbuf[0]=0.1*floor(sqrt(baseb_totpwr[baseb_pn]));
      baseb_pn=(baseb_pn+1)&baseband_mask;
      basebrawnet_pa=(basebrawnet_pa+4)&basebrawnet_mask;
      }
#endif


    }  
  if(kill_all_flag) goto da_output_error;
  if(thread_command_flag[THREAD_RX_OUTPUT] != THRFLAG_ACTIVE)goto stop_output;
  if(audio_dump_flag == 1 || rx_audio_out == -1)
    {
    while( ((daout_pa-daout_px+daout_size)&daout_bufmask) > 
                                                 snd[RXDA].block_bytes)
      {
      daout_px=(daout_px+snd[RXDA].block_bytes)&daout_bufmask;
      if(kill_all_flag) goto da_output_error;
      if(thread_command_flag[THREAD_RX_OUTPUT] != THRFLAG_ACTIVE)goto stop_output;
      }
    lir_sleep(sleep_time);
    if(kill_all_flag) goto da_output_error;
    if(thread_command_flag[THREAD_RX_OUTPUT] != THRFLAG_ACTIVE)goto stop_output;
    }
  while(daout_px != daout_py)
    {
    if(wav_write_flag != 0)
      {
      if(fwrite(&daout[daout_py],snd[RXDA].block_bytes,1,wav_file)!=1)
                                                        wavsave_start_stop(0);
      wavfile_bytes+=snd[RXDA].block_bytes;
      }
    da_block_counter+=1;
    daout_py=(daout_py+snd[RXDA].block_bytes)&daout_bufmask;
    if(kill_all_flag) goto da_output_error;
    if(thread_command_flag[THREAD_RX_OUTPUT] != THRFLAG_ACTIVE)goto stop_output;
    }
#if BUFBARS == TRUE
  if(baseband_handle != NULL && timinfo_flag!=0)
    {
    i=(daout_pa+2-daout_px+daout_bufmask+1)&daout_bufmask;
    i/=daout_indicator_block;
    if(i != rx_indicator_bufpos[7])
      {
      rx_indicator_bufpos[7]=i;
      lir_hline(indicator_first_pixel,indicator_ypix-21,
                          indicator_first_pixel+i-1,15);
      lir_hline(indicator_first_pixel+i,indicator_ypix-21,
                          indicator_first_pixel+INDICATOR_SIZE-1,1);
      }
    }
#endif
  }
stop_output:;                         
if(blocking)
  {
  lir_set_event(EVENT_BLOCKING_RXOUT);
  linrad_thread_stop_and_join(THREAD_BLOCKING_RXOUT);
  lir_close_event(EVENT_BLOCKING_RXOUT);
  }
//stop_output_2:;
daout_px=0;
daout_pa=0;
daout_py=0;
if(thread_command_flag[THREAD_RX_OUTPUT] == THRFLAG_IDLE)
  {
  if(rx_audio_out >= 0)lir_empty_da_device_buffer();
  thread_status_flag[THREAD_RX_OUTPUT] = THRFLAG_IDLE;
  lir_sleep(10000);
  while(thread_command_flag[THREAD_RX_OUTPUT] == THRFLAG_IDLE)
    {
    lir_sleep(5000);
    if(kill_all_flag) goto da_output_error;
    }
  daout_px=0;
  daout_pa=0;
  while(thread_command_flag[THREAD_RX_OUTPUT] == THRFLAG_SEMCLEAR)
    {
    thread_status_flag[THREAD_RX_OUTPUT] = THRFLAG_SEMCLEAR;
    lir_await_event(EVENT_RX_START_DA);
    if(kill_all_flag) goto da_output_error;
    }
  if(thread_command_flag[THREAD_RX_OUTPUT] == THRFLAG_ACTIVE)
    {
    if( rx_audio_out != -1 && (local_channels != rx_daout_channels ||
                               local_bytes != rx_daout_bytes ||
                               local_fft3_blocktime != fft3_blocktime))
      {
      sys_func(THRFLAG_CLOSE_RX_SNDOUT);
      lir_sched_yield();
      if(rx_daout_bytes == 1)
        {
        cc=0x80;
        }
      else
        {
        cc=0;
        }  
      daout_px=0;
      daout_pa=0;
      memset(&daout[daout_px],cc,snd[RXDA].block_bytes);
      sys_func(THRFLAG_OPEN_RX_SNDOUT);
      if(kill_all_flag) goto da_output_error;
      lir_empty_da_device_buffer();
      da_start_bytes=-1;
      local_channels=rx_daout_channels;
      local_bytes=rx_daout_bytes;
      local_fft3_blocktime=fft3_blocktime;
      }
    if(kill_all_flag) goto da_output_error;
    goto resume;
    }
  }
da_output_error:;
count_rx_underrun_flag=FALSE;
if(blocking && thread_command_flag[THREAD_BLOCKING_RXOUT] != THRFLAG_NOT_ACTIVE)
  {
  linrad_thread_stop_and_join(THREAD_BLOCKING_RXOUT);
  }
if(rx_audio_out >= 0)
  {
  if( (ui.use_alsa&(PORTAUDIO_RXIN+PORTAUDIO_RX_OUT)) == 0)
    {
    while( rx_audio_out == tx_audio_in)lir_sleep(10000);
    }
  sys_func(THRFLAG_CLOSE_RX_SNDOUT);
  }
thread_status_flag[THREAD_RX_OUTPUT]=THRFLAG_RETURNED;
while(thread_command_flag[THREAD_RX_OUTPUT] != THRFLAG_NOT_ACTIVE &&
      thread_command_flag[THREAD_RX_OUTPUT] != THRFLAG_KILL)
  {
  lir_sleep(1000);
  }
}

void make_audio_signal(void)
{
int ptr, i,j,i1,i2,i3,i4,mm,nn;
float t1,t2,t3,t4,t5,t6,t7;
double dt1;
short int *intvar;
float r1,r2,a1,a2,b1,b2;
float rdiff, final_gain1, final_gain2;
int loop_counter;
// Resample the baseband signal so we get the correct sampling speed
// for the D/A converter.
// For each interval in baseb_out there is a non integer number of points
// in da_output.
dt1=1/sqrt(rx_daout_sin*rx_daout_sin+rx_daout_cos*rx_daout_cos);
rx_daout_sin*=dt1;
rx_daout_cos*=dt1;
final_gain1=final_gain2=0;
mm=snd[RXDA].framesize;
nn=2*baseb_channels;
r1=daout_pa/(mm*da_resample_ratio);
loop_counter=0;


resamp:;
r2=(daout_pa+mm)/(mm*da_resample_ratio);
i2=baseb_fx+r1+.5;
if(r2>r1)
  {
  i3=baseb_fx+r2+.5;
  }
else
  {
  i3=baseb_fx+r2+.5+daout_size/da_resample_ratio;
  }  
i2&=baseband_mask;
i3&=baseband_mask;
if(abs(i3-i2) > 1)
  {
  i2=baseb_fx+(r1+r2)/2;
  i3=i2+1;
  i2&=baseband_mask;
  i3&=baseband_mask;
  }
else
  {
  if(i3==i2)
    {
    i2=baseb_fx+r1;
    i2&=baseband_mask;
    if(i3==i2)
      {
      i3=i2+1;
      i3&=baseband_mask;
      }
    }
  }        
i4=(i3+1)&baseband_mask;
baseb_py=i4;
i1=(i2+baseband_mask)&baseband_mask;
baseb_wts=baseb_pa-r2-baseb_fx;
if(baseb_wts<0)baseb_wts+=baseband_size;
loop_counter++;
if(loop_counter > 2*baseb_output_block)lir_sched_yield();
if( ((daout_pa-daout_px+daout_size)&daout_bufmask) > (daout_size>>1) ||
     ((baseb_pa-baseb_py+baseband_size)&baseband_mask) < baseb_min_block)
  {
  return;  
  }
rdiff=r1+baseb_fx-i2;
if(rdiff > baseband_size/2)
  {
  rdiff-=baseband_size;
  }
dt1=rx_daout_cos;
rx_daout_cos= rx_daout_cos*rx_daout_phstep_cos+rx_daout_sin*rx_daout_phstep_sin;
rx_daout_sin= rx_daout_sin*rx_daout_phstep_cos-dt1*rx_daout_phstep_sin;
// Use Lagrange's interpolation formula to fit a third degree
// polynomial to 4 points:
//  a1=-rdiff *   (rdiff-1)*(rdiff-2)*baseb_out[nn*i1]/6
//     +(rdiff+1)*(rdiff-1)*(rdiff-2)*baseb_out[nn*i2]/2
//     -(rdiff+1)* rdiff   *(rdiff-2)*baseb_out[nn*i3]/2
//     +(rdiff+1)* rdiff   *(rdiff-1)*baseb_out[nn*i4]/6; 
// Rewrite slightly to save a few multiplications - do not
// think the compiler is smart enough to do it for us.
t1=rdiff-1;
t2=rdiff-2;
t3=rdiff+1;
t4=t1*t2;
t5=t3*rdiff;
t6=rdiff*t4;
t4=t3*t4;
t7=t5*t2;
t5=t5*t1;
final_gain1=daout_gain;
final_gain2=daout_gain;
if(bg.agc_flag == 1)
  {
  if( daout_gain*baseb_agc_level[i2] > bg_agc_amplimit)
    {
    final_gain1=bg_agc_amplimit/baseb_agc_level[i2];
    final_gain2=final_gain1;
    }
  }    
if(bg.agc_flag == 2)
  {
  if( daout_gain*baseb_agc_level[2*i2] > bg_agc_amplimit)
    {
    final_gain1=bg_agc_amplimit/baseb_agc_level[2*i2];
    }
  if( daout_gain*baseb_agc_level[2*i2+1] > bg_agc_amplimit)
    {
    final_gain2=bg_agc_amplimit/baseb_agc_level[2*i2+1];
    }
  }    
if(baseb_channels == 1)
  {
  a1=final_gain1*(((t5*baseb_out[nn*i4  ]-t6*baseb_out[nn*i1  ])/3
                    +t4*baseb_out[nn*i2  ]-t7*baseb_out[nn*i3  ])/2);
  a2=final_gain1*(((t5*baseb_out[nn*i4+1]-t6*baseb_out[nn*i1+1])/3
                    +t4*baseb_out[nn*i2+1]-t7*baseb_out[nn*i3+1])/2);
  if( (ui.network_flag&NET_RXOUT_BASEB) != 0)
    {
    short int *ntbuf;
    ntbuf=(void*)&baseb_netsend_buffer[basebnet_pa];
    if(rx_daout_bytes == 1)
      {
      ntbuf[0]=floor(256*a2);
      ntbuf[1]=floor(256*a1);
      }
    else
      {
      ntbuf[0]=floor(a2);
      ntbuf[1]=floor(a1);        
      }
    basebnet_pa=(basebnet_pa+4)&basebnet_mask;
    }
  if(rx_mode == MODE_FM)
    {
    if(a1 < -bg_amplimit)a1=-bg_amplimit;
    if(a1 > bg_amplimit)a1=bg_amplimit;
    if(bg_maxamp < fabs(a1))bg_maxamp=fabs(a1); 
    if(rx_daout_channels == 2)
      {
      if(bg_maxamp < fabs(a2))bg_maxamp=fabs(a2); 
      if(a2 < -bg_amplimit)a2=-bg_amplimit;
      if(a2 > bg_amplimit)a2=bg_amplimit;
      }
    if(rx_daout_bytes == 1)
      {
      daout[daout_pa]=0x80+a1;
      if(rx_daout_channels == 2)daout[daout_pa+1]=0x80+a2;
      }
    else
      {
      i=a1;
      intvar=(void*)(&daout[daout_pa]);
      intvar[0]=i;
      i=a2;
      if(rx_daout_channels == 2)intvar[1]=i;
      }  
    }
  else
    {
    if(bg_expand != 2)
      {
      t1=sqrt(a1*a1+a2*a2);
      if(t1 > 0.5)
        {
        if(t1<bg_amplimit)t2=t1; else t2=bg_amplimit;
        if(bg_expand == 1)t2=bg_expand_a*(exp(bg_expand_b*t2)-1);
        if(bg_maxamp < t2)bg_maxamp=t2; 
        t2/=t1;
        t2*=a1*rx_daout_cos+a2*rx_daout_sin;
        }
      else
        {
        t2=0;
        }  
      if(rx_daout_bytes == 1)
        {
        daout[daout_pa]=0x80+t2;
        if(rx_daout_channels == 2)daout[daout_pa+1]=0x80+t2*da_ch2_sign;
        }
      else
        {
        i=t2;
        intvar=(void*)(&daout[daout_pa]);
        intvar[0]=i;
        if(rx_daout_channels == 2)intvar[1]=i*da_ch2_sign;
        }  
      }
    else
      {
      t2=a1*rx_daout_cos+a2*rx_daout_sin;
      if(t2 > bg_amplimit)t2=bg_amplimit;
      if(t2 < -bg_amplimit)t2=-bg_amplimit;
      if(bg_maxamp < t2)bg_maxamp=t2; 
      if(rx_daout_bytes == 1)
        {
        daout[daout_pa]=0x80+t2;
        if(rx_daout_channels == 2)daout[daout_pa+1]=0x80+t2*da_ch2_sign;
        }
      else
        {
        i=t2;
        intvar=(void*)(&daout[daout_pa]);
        intvar[0]=i;
        if(rx_daout_channels == 2)intvar[1]=i*da_ch2_sign;
        }  
      }
    }
  }
else
  {
  a1=final_gain1*(((t5*baseb_out[nn*i4  ]-t6*baseb_out[nn*i1  ])/3
                    +t4*baseb_out[nn*i2  ]-t7*baseb_out[nn*i3  ])/2);
  a2=final_gain1*(((t5*baseb_out[nn*i4+1]-t6*baseb_out[nn*i1+1])/3
                    +t4*baseb_out[nn*i2+1]-t7*baseb_out[nn*i3+1])/2);
  b1=final_gain2*(((t5*baseb_out[nn*i4+2]-t6*baseb_out[nn*i1+2])/3
                    +t4*baseb_out[nn*i2+2]-t7*baseb_out[nn*i3+2])/2);
  b2=final_gain2*(((t5*baseb_out[nn*i4+3]-t6*baseb_out[nn*i1+3])/3
                    +t4*baseb_out[nn*i2+3]-t7*baseb_out[nn*i3+3])/2);
  if( (ui.network_flag&NET_RXOUT_BASEB) != 0)
    {
    short int *ntbuf;
    ntbuf=(void*)&baseb_netsend_buffer[basebnet_pa];
    if(rx_daout_bytes == 1)
      {
      ntbuf[0]=floor(256*a2);
      ntbuf[1]=floor(256*a1);
      ntbuf[2]=floor(256*b2);
      ntbuf[3]=floor(256*b1);
      }
    else
      {
      ntbuf[0]=floor(a2);
      ntbuf[1]=floor(a1);        
      ntbuf[2]=floor(b2);
      ntbuf[3]=floor(b1);        
      }
    basebnet_pa=(basebnet_pa+8)&basebnet_mask;
    }
  if(bg_expand != 2 && bg.agc_flag != 2)
    {
    t1=a1*a1+a2*a2;
    t2=b1*b1+b2*b2;
    if(t1 < t2)t1=t2;
    t1=sqrt(t1);
    if(t1 > 0.5)
      {
      if(t1<bg_amplimit)t2=t1; else t2=bg_amplimit;
      if(bg_expand == 1)t2=bg_expand_a*(exp(bg_expand_b*t2)-1);
      if(bg_maxamp < t2)bg_maxamp=t2; 
      t2/=t1;
      t1=t2*(a1*rx_daout_cos+a2*rx_daout_sin);
      t2=t2*(b1*rx_daout_cos+b2*rx_daout_sin);
      }
    else
      {
      t1=0;
      t2=0;
      }  
    if(rx_daout_bytes == 1)
      {
      daout[daout_pa]=0x80+t1;
      daout[daout_pa+1]=0x80+t2;
      }
    else
      {
      intvar=(void*)(&daout[daout_pa]);
      intvar[0]=t1;
      intvar[1]=t2;
      }  
    }
  else
    {
    t1=a1*rx_daout_cos+a2*rx_daout_sin;
    t2=b1*rx_daout_cos+b2*rx_daout_sin;
    if(t1 > bg_amplimit)t1=bg_amplimit;
    if(t1 < -bg_amplimit)t1=-bg_amplimit;
    if(t2 > bg_amplimit)t2=bg_amplimit;
    if(t2 < -bg_amplimit)t2=-bg_amplimit;
    if(bg_maxamp < t1)bg_maxamp=t1; 
    if(bg_maxamp < t2)bg_maxamp=t2; 
    if(rx_daout_bytes == 1)
      {
      daout[daout_pa]=0x80+t1;
      daout[daout_pa+1]=0x80+t2;
      }
    else
      {
      intvar=(void*)(&daout[daout_pa]);
      intvar[0]=t1;
      intvar[1]=t2;
      }  
    }
  }
r1=r2;
ptr=(daout_pa-mm*daout_upsamp+daout_size)&daout_bufmask;
if(daout_upsamp==1)
  {
  daout_pa=(daout_pa+mm)&daout_bufmask;
  }
else
  {
  if(rx_daout_channels == 2)
    {
    if(rx_daout_bytes == 1)
      {
// Two channels, one byte.        
      t1=(float)((char)(daout[daout_pa]-0x80)-(char)(daout[ptr]-0x80))/daout_upsamp;
      t2=(char)(daout[ptr]-0x80);
      t3=(float)((char)(daout[daout_pa+1]-0x80)-(char)(daout[ptr+1]-0x80))/daout_upsamp;
      t4=(char)(daout[ptr+1]-0x80);
      for(i=1; i<daout_upsamp; i++)
        {
        t2+=t1;
        t4+=t3;
        daout[ptr+2*i  ]=0x80+t2;
        daout[ptr+2*i+1]=0x80+t4;
        }
      daout_pa=(daout_pa+2*daout_upsamp)&daout_bufmask;
      }
    else
      {
// Two channels, two bytes.        
      intvar=(void*)(&daout[daout_pa]);
      i=intvar[0];
      j=intvar[1];
      intvar=(void*)(&daout[ptr]);
      t1=(float)(i-intvar[0])/daout_upsamp;
      t2=intvar[0];
      t3=(float)(j-intvar[1])/daout_upsamp;
      t4=intvar[1];
      for(i=1; i<daout_upsamp; i++)
        {
        t2+=t1;
        t4+=t3;
        intvar[2*i  ]=t2;
        intvar[2*i+1]=t4;
        }
      daout_pa=(daout_pa+4*daout_upsamp)&daout_bufmask;
      }  
    } 
  else
    {
    if(rx_daout_bytes == 1)
      {
// One channel, one byte.        
      t1=(float)((char)(daout[daout_pa]-0x80)-(char)(daout[ptr]-0x80))/daout_upsamp;
      t2=(char)(daout[ptr]-0x80);
      for(i=1; i<daout_upsamp; i++)
        {
        t2+=t1;
        daout[ptr+i]=0x80+t2;
        }
      daout_pa=(daout_pa+daout_upsamp)&daout_bufmask;
      }  
    else
      {
// One channel, two bytes.        
      intvar=(void*)(&daout[daout_pa]);
      i=intvar[0];
      intvar=(void*)(&daout[ptr]);
      t1=(float)(i-intvar[0])/daout_upsamp;
      t2=intvar[0];
      for(i=1; i<daout_upsamp; i++)
        {
        t2+=t1;
        intvar[i]=t2;
        }
      daout_pa=(daout_pa+2*daout_upsamp)&daout_bufmask;
      }  
    }
  }    
if(daout_pa > daout_bufmask)lirerr(994578);
if(daout_pa == 0)
  {
  baseb_fx+=r2;
  r1=0; 
  da_resample_ratio=new_da_resample_ratio;
  if(baseb_fx>baseband_size)baseb_fx-=baseband_size;
  if(use_bfo != 0)
    {
    t1=(2*PI_L*bg.bfo_freq*daout_upsamp)/genparm[DA_OUTPUT_SPEED];
    rx_daout_phstep_cos=cos(t1);
    rx_daout_phstep_sin=sin(t1);
    }
  }
if(thread_command_flag[THREAD_RX_OUTPUT] != THRFLAG_ACTIVE)return;
goto resamp;  
}
